/*
 * Naui - Not a UI
 *
 * Code attempts to print some basic diagnostic on an Intel Galileo board using
 * mostly information from sysfs.  Using this info it uses libxml2 to create an
 * XHTML webpage where the information can be displayed.  It is meant to
 * 'create' this webpage from a template on boot and then update it every 'x'
 * seconds when called.
 *
 * Under MIT license. See COPYING.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gnu/libc-version.h>
#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <sys/utsname.h>

#define STATPAGE "/www/pages/index.html"
#define BASEPAGE "/usr/share/naui/index.html"
#define ENCODING "UTF-8"
#define ITEMTAG "div"
#define MAX_GCC_STRING 8
#define MAX_XML_FIND_STRING 128
#define ITEMATTR "name"

#define GALILEO 1
#define VERBOSE 0

xmlDocPtr doc = NULL;
xmlXPathContext *xpathCtx;

unsigned int
grab_wfile(char **arg, char *file)
{
  FILE *cmdline = fopen(file, "rb");
  if (cmdline == NULL) {
    fprintf(stderr, "Could not open file: %s\n", file);
    return 1;
  }
  size_t size = 0;
  while (getdelim(arg, &size, 0, cmdline) != -1);
  fclose(cmdline);

  return 0;
}

unsigned int
grab_uname(char **str)
{
  struct utsname unameD;
  uname(&unameD);
  unsigned int len = strlen(unameD.release) + 1;
  *str = (char *) malloc(sizeof(char) * len);
  snprintf(*str, len, "%s", unameD.release);
  return 0;
}
unsigned int
grab_clock_speed(char **str)
{
  char *raw = NULL;
  if (grab_wfile(&raw, "/proc/cpuinfo"))
    return 1;
  char *w = strstr(raw, "cpu MHz");
  if (w == NULL) {
    fprintf(stderr, "Looks like your kernel needs some work!\n");
    return 1;
  }
  char buffer[12];
  sscanf(w, "cpu MHz         : %12s", buffer);
  int len = strlen(buffer);
  *str = (char *) malloc(sizeof(char) * len);
  snprintf(*str, len, "%s", buffer);
  return 0;
}

unsigned int
grab_cpu_flag(char **str)
{
  char *raw = NULL;
  if (grab_wfile(&raw, "/proc/cpuinfo"))
    return 1;
  char *w = strstr(raw, "flags");
  if (w == NULL) {
    fprintf(stderr, "Looks like your kernel needs some work!\n");
    return 1;
  }
  w = strtok(w, "\n");
  char buffer[512];
  sscanf(w, "flags           : %500c", buffer);
  int len = strlen(buffer);
  *str = (char *) malloc(sizeof(char) * len);
  snprintf(*str, len, "%s", buffer);
  return 0;
}

unsigned int
grab_temp(char **str)
{
  char *raw = NULL;
  grab_wfile(&raw, "/sys/class/thermal/thermal_zone0/temp");
  if (raw == NULL)
    return 1;
  int len = strlen(raw) - 3;
  *str = (char *) malloc(sizeof(char) * (len));
  snprintf(*str, (len), "%s", raw);
  return 0;
}
xmlNode
*create_node(const char* id, const char* name, const char* value)
{
  xmlNode *node = xmlNewNode(NULL, BAD_CAST ITEMTAG);
  xmlNewProp(node, BAD_CAST ITEMATTR, BAD_CAST id);
  xmlNodeSetContent(node, BAD_CAST name);
  xmlNodeAddContent(node, BAD_CAST " : "); // Let libxml do the concat of content
  xmlNodeAddContent(node, BAD_CAST value);
  return node;
}

unsigned int
find_node(xmlNode **nd, const char* id)
{
  char search[MAX_XML_FIND_STRING];
  snprintf(search, MAX_XML_FIND_STRING,"/html/body/section[@id='varible']/%s[@%s='%s']",ITEMTAG ,ITEMATTR, id);
  xmlXPathObject *path_dyn = xmlXPathEvalExpression((xmlChar*)search, xpathCtx);
  *nd = path_dyn->nodesetval->nodeTab[0];
  if(*nd == NULL) {
    fprintf(stderr, "Could not find xmlNode:\n%s\n", search);
    return 0;
  }
  return 1;
}

unsigned int
update_node(xmlNode **node, const char* name, const char* value)
{
  xmlNodeSetContent(*node, BAD_CAST name);
  xmlNodeAddContent(*node, BAD_CAST " : ");
  xmlNodeAddContent(*node, BAD_CAST value);
  return 1;
}

/*
 * Wrapper around xmlParseFile that makes sure doc is properly closed
 */
static unsigned int
open_xml_page(char *page)
{
  if (doc != NULL)
    xmlFreeDoc(doc);

  doc = xmlParseFile(page);
  if (doc == NULL)
    return 0;
  else
    xpathCtx = xmlXPathNewContext(doc);

  return 1;
}
/*
 * Call to write only info that may change after boot
 * temp, cpu load, etc...
 */
static unsigned int
update_rebase(int init)
{
  if (!init && !open_xml_page(STATPAGE)) {
    printf("ERROR: Unable to load existing naui page. Run without any flags!\n");
    return 0;
  }
  xmlXPathObject *path_dyn = xmlXPathEvalExpression((xmlChar*)"/html/body/section[@id='varible']", xpathCtx);
  xmlNode *node_dyn = path_dyn->nodesetval->nodeTab[0];

  xmlNode *cpu_avg = NULL;
  xmlNode *temp = NULL;

  if(init) {
    cpu_avg = xmlAddChild(node_dyn, create_node("cpu_avg", "CPU Load", NULL));
    temp = xmlAddChild(node_dyn, create_node("temp", "Temperature", NULL));
  }
  if(cpu_avg == NULL)
    find_node(&cpu_avg, "cpu_avg");
  char *cpu_avg_str = NULL;
  grab_wfile(&cpu_avg_str, "/proc/loadavg");
  update_node(&cpu_avg, "CPU Load", cpu_avg_str);
 // free(cpu_avg_str);
  if(temp == NULL)
    find_node(&temp, "temp");  free(cpu_avg_str);
  char *temp_str = NULL;
  //CHANGE TO ACTUALY GALLIEO TEMP READING.
  grab_temp(&temp_str);
  update_node(&temp, "Temperature (celsius)", temp_str);
 // free(temp_str);

  return 1;
}

static unsigned int
initial_write()
{
  if (!open_xml_page(BASEPAGE))
    return 0;
  xmlXPathObject *head_p = xmlXPathEvalExpression((xmlChar*)"/html/body/h1", xpathCtx);
  xmlNode *header = head_p->nodesetval->nodeTab[0];
  char *issue_raw = NULL;
  grab_wfile(&issue_raw, "/etc/issue");
  int issue_len = strlen(issue_raw) - 6;
  char issue[issue_len];
  snprintf(issue, issue_len, "%s", issue_raw);
  xmlNodeSetContent(header, BAD_CAST issue);

  xmlXPathObject *path_init = xmlXPathEvalExpression((xmlChar*)"/html/body/section[@id='init']", xpathCtx);
  xmlNode *node_init = path_init->nodesetval->nodeTab[0];

  char *revision_str = NULL;
  grab_wfile(&revision_str,"/etc/version");
  xmlAddChild(node_init, create_node("version","OS Version", revision_str));
  free(revision_str);

  char *kern_version = NULL;
  grab_uname(&kern_version);
  xmlAddChild(node_init, create_node("kernelver", "Kernel Version", kern_version));
  free(kern_version);

  char *flash_ver = NULL;
  grab_wfile(&flash_ver,"/sys/firmware/board_data/flash_version");
  xmlAddChild(node_init, create_node("flashver","Flash Version", flash_ver));
  free(flash_ver);

  char gcc_v[MAX_GCC_STRING];
  snprintf(gcc_v, MAX_GCC_STRING, "%d.%d", __GNUC__, __GNUC_MINOR__);
  xmlAddChild(node_init, create_node("cc", "GNU cc", gcc_v));

  xmlAddChild(node_init, create_node("libc", "eglibc", gnu_get_libc_version()));


  char *mac_one = NULL;
  if (grab_wfile(&mac_one,"/sys/class/net/enp0s20f6/address") == 1) {
    grab_wfile(&mac_one,"/sys/class/net/eth0/address");
  }
  xmlAddChild(node_init, create_node("mac1", "MAC Address 1", mac_one));
  free(mac_one);
#if VERBOSE
  char *mac_two = NULL;
  grab_wfile(&mac_two,"/sys/firmware/board_data/2nd MAC");
  xmlAddChild(node_init, create_node("mac2", "MAC Address 2", mac_two));
  free(mac_two);

  char *cpu_flag_str = NULL;
  grab_cpu_flag(&cpu_flag_str);
  xmlAddChild(node_init, create_node("cpu_flag","CPU Flags", cpu_flag_str));
#endif
  char *clock_str = NULL;
  grab_clock_speed(&clock_str);
  xmlAddChild(node_init, create_node("cpu_clock","CPU Clock", clock_str));

  // Done ith bootime. Going for initial dynamic items.
  update_rebase(1);

   return 1;
}


int
main(int argc, char *argv[])
{
  if (argc > 1) {
    update_rebase(0);
  } else {
    initial_write();
  }
  xmlSaveFileEnc(STATPAGE, doc, ENCODING);
  xmlFreeDoc(doc);
  exit(EXIT_SUCCESS);
}
